#!/bin/sh
#
# ezshaper 1.0rc    (C) Jannes Faber 2002, 2003
# 
# Makes configuring traffic shaping on a linux box easy. Well a little less
# complicated, for me anyway. 
#
# Requirements:
# - Linux 2.4.xx with HTB (2.4.20+ or so have HTB built in) and iptables.
# - A tc version capable of using HTB. (It may be possible to adapt this
#       script for other qdiscs than HTB).
# - Use of the iptables MARK target (only a few numbers). If you are already
#   using MARK for other purposes, you might want to watch out for conflicts.
#   See the OFFSET_PRIO variable to change the MARK numbers used by ezshaper.
# 
# How it works:
# I wrote this script for the following basic setup: 
#   LAN <---> linux box running firewall+NAT <---> Internet
# but it should also work on stand-alone machines connected to the internet
# and some other "simple" situations.
# 
# EzShaper will insert 1 rule in both your OUTPUT and FORWARD chains of 
# iptables and creates its own chain (EZSHAPER by default). It won't mess up 
# any of your normal firewall rules and will clean itself up nicely when you
# use "ezshaper stop". 
# 
# The EZSHAPER chain is filled with MARK rules, which you can easily specify
# in the ezshaper.cfg file.
# 
# EzShaper will then set up all the necessary tc rules again using the info
# from the config file.
# 
# This will simply result in a htb qdisc with a few classes, each with their
# own priority, rate and ceil settings.
# 
# Use "ezshaper status" for some info on the current configuration. Use 
# "ezshaper -h" for some handy info.
# 
# I use tab size 2.
# 
# Thanks go to:
# - Martin Devera for HTB
#   http://luxik.cdi.cz/~devik/qos/htb
# - The LARTC HOWTO and mailing list
#   http://www.lartc.org
# - Stef Coene
#   http://www.docum.org
# 
# Wishlist / TODO @@@
# - Make setting r2q and/or quantum stuff changable in config file.
# - De-couple prio and classid, prio should be a setting just like rate/ceil/r2q
#   (allow several classes to have the same prio) Allow sub-classes: parent should
#   also be a setting for each class.
# - Fix MTU stuff (currently messes up some configurations)
# - When using TTL target: only exclude traceroute icmp, not all icmp
# - Allow different rules for different intefaces (through multiple cfg files?)
#   Mainly for QOS for VPNs (ipsec) and for shaping incoming traffic.
# - Allow using the same ruleset on two or more interfaces at once (allow 
#   IF_LAN containing 1 or more interfaces).
# - Use the iptables CONNMARK target to make things more efficient for iptables
#   (not a really big issue for normal situations, but would be nicer)
# - Maybe duplicate all MARK rules with -j RETURN so that iptables counters
#   are correct (or use mark match to see if prev rule set a mark).
# - Implement not sign ( ! ) in cfg file

####################
# Defaults
####################
defaults()
{
  # Default configuration file, you can seperate alternatives by a space,
  # or leave it empty. A .cfg file with the same name and directory as this 
  # script will be used if none of the ones below can be found.
  CONFIGFILE=/etc/ezshaper.cfg
  
  # Interfaces
  IF_INET=    # Specify your internet interface or leave empty for auto-detect
  IF_LAN=     # Not used for nowm just leave it empty
  
  # IP address
  IP_LAN=     # The IP of your internet interface or leave empty for auto-detect
  
  # Executables
  IPTABLES="iptables"
  TC="tc"
  IP="ip"                 # Only used for autodetecting IF_INET (and doing MTU changes @@@)
  IFCONFIG="ifconfig"     # Only used for autodetecting IP_INET
  
  # Add this to PRIO (tc class id and iptables MARK can not be 0 !!)
  OFFSET_PRIO=10
  
  # Changing the qdisc very likely needs additional changes to the script!!
  QDISC="htb"
  
  # Names of the iptables chains to use
  CHAIN_INP=INPUT
  CHAIN_OUT=OUTPUT
  CHAIN_PRE=PREROUTING
  CHAIN_POS=POSTROUTING
  CHAIN_FOR=FORWARD
  CHAIN_EZ=EZSHAPER     # The name of the custom iptables chain
  
  # Version stuff
  VERSION="1.0rc"
  VERSIONDATE="2003-10-09"
  
  # What to do (these variables are set using command line options, no need
  # to change them here)
  DEBUGMODE=0
  VERBOSE=0
  QUIET=0
  DO_ALL=1
  DO_IPTABLES=0
  DO_TC=0
  DO_WATCH=0
}

####################
# Misc functions
####################
# Report an error and exit
error()
{
  echo -e "ERROR:\t$*" >&2
  exit 1
}

# Just a warning
warning()
{
  [ $QUIET = 0 ] && echo -e "WARNING: $*"
}

# Execute a shell command, or not when in debugmode.
# And display the command (when in verbose mode)
run()
{
  [ $VERBOSE   = 1 ] && echo $*
  [ $DEBUGMODE = 0 ] && $*
}

# Add a kilo, Mega or Giga suffix to a number
addsuffix()
{
  [ $# != 2 ] && return
  local -i VAL=$1
  local -i BASE=$2
  local SUFFIX=
  for SUFFIX in . k M G
  do
    [ $VAL -lt $BASE ] && break
    VAL=$[$VAL / $BASE]
  done
  [ "$SUFFIX" = "." ] && SUFFIX=
  [ $VAL -gt 0 ] && echo -n "$VAL$SUFFIX"
}

# Find the config file
find_config()
{
  local FILES=$CONFIGFILE
  local FILE=
  CONFIGFILE=
  
  for FILE in $FILES
  do
    if [ -f "$FILE" ]; then
      CONFIGFILE=$FILE
      break
    fi
  done
  [ "$CONFIGFILE" = "" ] && error "Config file not found, tried:\n $FILES"
}

find_if_inet()
{
  local ROUTE=`$IP route list match 0.0.0.0`
  [ "$ROUTE" = "" ] && return
  set $ROUTE
  while [ "$1" != "dev" -a "$1" != "" ]
  do
    shift
  done
  IF_INET="$2"
  [ "$IF_INET" = "" ] && error "Could not find the interface connected to the internet. Maybe your connection is down?"
}

# Find the ip of IF_INET and put it in IP_INET (only one supported for now @@@)
find_ip_inet()
{
  for WORD in $($IFCONFIG $IF_INET)
  do
    [ "${WORD:0:5}" = "addr:" ] && IP_INET=${WORD:5} && break
  done
}

# Get an array of DNS servers from /etc/resolve.conf
find_dns()
{
  [ ! -f /etc/resolv.conf ] && return
  while read LINE
  do
    [ "$LINE" = "" ] && continue
    set $LINE
    [ "$1" = "nameserver" ] && DNS[${#DNS[*]}]=$2
  done < /etc/resolv.conf
}

# Translate "ME" and "DNSx" from config file into IP address
keyword2ip()
{
  if [ ${#1} = 4 -a "${1:0:3}" = "DNS" ]; then
    echo ${DNS[${1:3:1}]}
  elif [ "$1" = "ME" ]; then
    echo $IP_INET
  else
    echo $1
  fi
}

calc_rate()
{
  local PERC=${1%%%}
  if [ "$PERC" != "$1" ]; then
    echo $[$LINK_RATE_VAL * $PERC / 100]$LINK_RATE_UNIT
  else
    echo $1
  fi
}

show_header()
{
  echo "# EzShaper $VERSION $VERSIONDATE"
}

show_config()
{
  echo "# Using config file: $CONFIGFILE"
  [ "$IF_INET" != "" ] && echo "# Internet interface: $IF_INET"
  echo
}

show_usage()
{
  echo "Usage: ${0##*/} [-${OPTIONS/:/}] {start|stop|restart|status}"
}

####################
# Parse options
####################
# Parse the command line options passed to this script.
parse_options()
{
  local FILE=${0%.sh}.cfg
  CONFIGFILE="$CONFIGFILE $FILE ./${FILE##*/}"
  
  # Use : after an option to let it have an argument passed with it (goes into $OPTARG).
  OPTIONS="dhitvq"
  while getopts ":$OPTIONS" OPTION
  do
    case $OPTION in
      h )
          show_header
          show_usage
          echo
          echo "Note that any of the following options MUST be specified BEFORE the command!"
          echo "-d  debugmode, only displays commands to be executed"
          echo "-h  this help"
          echo "-i  do iptables part only"
          echo "-t  do tc part only"
          echo "-v  verbose (show commands to be executed/extra status details)"
          echo "-q  quiet (suppress warnings)"
          echo
          echo "Config file is the first of these found:"
          echo -e "\t${CONFIGFILE// /\\n\\t}"
          find_config
          echo "Currently using:"
          echo -e "\t$CONFIGFILE"
          return
          ;;
      d ) # debugmode: display command that would have been executed
          DEBUGMODE=1
          VERBOSE=1
          ;;
      i ) # Do the Iptables part
          DO_IPTABLES=1
          DO_ALL=0
          ;;
      t ) # Do the TC part
          DO_TC=1
          DO_ALL=0
          ;;
      v ) # Be verbose
          VERBOSE=1
          ;;
      q ) # Be quiet (no warnings)
          QUIET=1
          ;;
    esac
  done
  shift $(($OPTIND - 1))
  # Move argument pointer to next.
  
  if [ $DO_ALL = 1 ]; then
    DO_IPTABLES=1
    DO_TC=1
  fi
  
  find_config
  [ "$IF_INET" = "" ] && find_if_inet
  [ "$IP_INET" = "" ] && find_ip_inet
  
  case "$1" in
    start)
      [ $VERBOSE = 1 ] && show_header && show_config
      command_start
      ;;
    stop)
      [ $VERBOSE = 1 ] && show_header && show_config
      command_stop
      ;;
    restart)
      [ $VERBOSE = 1 ] && show_header && show_config
      command_stop
      command_start
      ;;
    status)
      show_header
      show_config
      command_status
      ;;
    *)
      show_header
      show_usage
      exit 1
      ;;
  esac
}

####################
# Command stop
####################
command_stop()
{
  [ $DO_IPTABLES = 1 ] && iptables_stop
  [ $DO_TC       = 1 ] && tc_stop
}

####################
# Command start
####################
command_start()
{
  declare -a CLASSES=       # Collect all classes used
  declare -i DEFAULTCLASS=0 # Collect all classes used
  declare LINK_RATE=        # Get link rate from config
  declare LINK_RATE_VAL=    # The value
  declare LINK_RATE_UNIT=   # The unit (kbit)
  declare -a DNS=           # Hold all DNS servers
  find_dns
  iptables_start            # Will parse the config file
  [ $DO_TC       = 1 ] && tc_start
}

####################
# Command status
####################
command_status()
{
  [ $DO_IPTABLES = 1 ] && iptables_status
  [ $DO_TC       = 1 ] && tc_status
}

####################
# iptables commands
####################
iptables_status()
{
  local OPTION=
  [ $VERBOSE = 1 ] && OPTION=" -v"
  run $IPTABLES --list $CHAIN_EZ -t mangle -n$OPTION
}

iptables_stop()
{
  CHAIN=`$IPTABLES --list $CHAIN_EZ -t mangle -n 2>/dev/null`
  if [ "$CHAIN" != "" ]; then
    #run $IPTABLES --delete $CHAIN_INP -t mangle -j $CHAIN_EZ
    run $IPTABLES --delete $CHAIN_OUT -t mangle -j $CHAIN_EZ
    run $IPTABLES --delete $CHAIN_FOR -t mangle -j $CHAIN_EZ
    
    run $IPTABLES --flush $CHAIN_EZ  -t mangle
    run $IPTABLES --delete-chain $CHAIN_EZ  -t mangle
  fi
}

iptables_start()
{
  CHAIN=`$IPTABLES --list $CHAIN_EZ -t mangle -n 2>/dev/null`
  if [ "$CHAIN" != "" -a $DO_IPTABLES = 1 -a $DEBUGMODE = 0 ]; then
    error Chain $CHAIN_EZ already exists, cannot start, aborting!
  else
    # Create ezshaper chain and jump to it
    if [ $DO_IPTABLES = 1 ]; then
      run $IPTABLES --new $CHAIN_EZ  -t mangle

      # run $IPTABLES --insert $CHAIN_INP -t mangle -j $CHAIN_EZ
      run $IPTABLES --insert $CHAIN_OUT -t mangle -j $CHAIN_EZ
      run $IPTABLES --insert $CHAIN_FOR -t mangle -j $CHAIN_EZ
    fi
    
    # Read config file and generate iptables rules
    exec 3< $CONFIGFILE  # Open "File" and assign fd 3 to it.
    
    local -i L=0          # Line number (for error messages)
    local LINE=           # Current line (array)
    local TARGET=         # Current target
    local TARGETX=        # Extra options for current target
    local PRIO=           # Current PRIO
    local PROTSUFFIX=     # Current suffix for protocol match
    local -i MIRRORRULE=0 # Target with this set works both ways: in and out
    local -i NESTCOUNT=0  # Current tc class id
    local -i CLASS=       # Current tc class id
    
    while read LINE <&3
    do
      L=$[$L+1]
      LINE=${LINE%%#*}    # Strip comments (#)
      [ "$LINE" = "" ] && continue
      set $LINE
      [ "$1" = "" ] && continue
      [ "$1" = "EXIT" ] && break
      if [ $NESTCOUNT = 0 ]; then     # Not inside a target block yet
        case $1 in
# @@@ Messes up routing table on some systems...
#         MTU)
#           if [ $2 -ge 100 -a $2 -le 1500 ]; then
#             [ $DO_TC = 1 ] && run $IP route change default dev $IF_INET mtu $2
#           else
#             error "Specified MTU ($2) out of range [100-1500]"
#           fi
#           ;;
          LINK)
            LINK_RATE=$2
            LINK_RATE_VAL=${LINK_RATE//[^0123456789]/}
            LINK_RATE_UNIT=${LINK_RATE//[0123456789]/}
            ;;
          MARK)
            PRIO=$2
            if [ "$PRIO" = "" ]; then
              error "No class id specified for MARK in line $L !"
            else
              CLASS=$[$PRIO+$OFFSET_PRIO]
              TARGET="MARK --set-mark $CLASS"
              shift 2
              if [ "$1" != "" ]; then
                RATE=`calc_rate $1`
                CEIL=`calc_rate $2`
                CLASSES[${#CLASSES[*]}]="$CLASS $RATE $CEIL"
                shift 2
                if [ "$1" = "DEFAULT" ]; then
                  shift
                  DEFAULTCLASS=$[${#CLASSES[*]}-1]
                fi
                TARGETX=$*
              fi
            fi
            ;;
          DROP)
            TARGET="DROP"
            shift 1
            TARGETX=$*
            MIRRORRULE=1    # This rule works both ways: in and out
            ;;
          TCPMSS)
            TARGET="TCPMSS --set-mss $2"
            shift 2
            TARGETX=$*
            PROTSUFFIX="--tcp-flags SYN,RST SYN"
            MIRRORRULE=1    # This rule works both ways: in and out
            ;;
          TTL)
            TARGET="TTL --ttl-set $2"
            shift 2
            TARGETX=$*
            ;;
          TOS)
            TARGET="TOS --set-tos $2"
            shift 2
            TARGETX=$*
            MIRRORRULE=1    # This rule works both ways: in and out
            ;;
          ECN)
            TARGET="ECN --ecn-tcp-remove"
            shift
            TARGETX=$*
            ;;
          {)                # Start target block
            NESTCOUNT=1
            ;;
          *)
            error "\nNot a valid target $1 in line #$L:\n $LINE"
            ;;
        esac
      elif [ "$1" = "}" ]; then # End of the target block
        NESTCOUNT=0
        TARGET=""
        MIRRORRULE=0
        PROTSUFFIX=""
        PRIO=""
      else                              # inside a target block
        local PROT=$1
        local SIP=$2
        local SPT=$3
        local DIP=$4
        local DPT=$5
        if [ $# -gt 5 ]; then
          shift 5
        else
          shift $#
        fi
        local EXTRA="$*"

        local RULE1=""
        local RULE2=""  # = RULE1 mirrored
        local RULE3=""
        local RULE4=""  # = RULE3 mirrored
        local MULTIPORT=""

        if [ "$SIP" != "" -a "${SIP:0:1}" != "0" -a "${SIP:0:3}" != "any" ]; then
          SIP=`keyword2ip $SIP`
          [ "$SIP" = "" ] && continue
          RULE1="$RULE1 --source $SIP"
          RULE2="$RULE2 --destination $SIP"
        fi

        if [ "$SPT" != "" -a "${SPT:0:1}" != "0" -a "${SPT:0:3}" != "any" ]; then
          RULE1="$RULE1 --source-port $SPT"
          RULE2="$RULE2 --destination-port $SPT"

          if [ "${SPT#*,}" != "$SPT" ]; then
            MULTIPORT="-m multiport"
          fi
        fi

        if [ "$DIP" != "" -a "${DIP:0:1}" != "0" -a "${DIP:0:3}" != "any" ]; then
          DIP=`keyword2ip $DIP`
          [ "$DIP" = "" ] && continue
          RULE1="$RULE1 --destination $DIP"
          RULE2="$RULE2 --source $DIP"
        fi

        if [ "$DPT" != "" -a "${DPT:0:1}" != "0" -a "${DPT:0:3}" != "any" ]; then
          RULE1="$RULE1 --destination-port $DPT"
          RULE2="$RULE2 --source-port $DPT"

          if [ "${DPT#*,}" != "$DPT" ]; then
            MULTIPORT="-m multiport"
          fi
        fi

        if [ "$EXTRA" != "" -a "${EXTRA:0:1}" != "#" ]; then
          RULE1="$RULE1 $EXTRA"
          RULE2="$RULE2 $EXTRA"
        fi

        if [ "$PROT" = "any" -o "$PROT" = "all" ]; then
          if [ "${TARGET:0:3}" = "TTL" ]; then
            PROT="-p ! ICMP"
          else
            PROT=""
          fi
        else
          PROT="-p $PROT"
        fi

        RULE1="-j $TARGET $PROT $PROTSUFFIX $MULTIPORT $RULE1 $TARGETX"
        RULE2="-j $TARGET $PROT $PROTSUFFIX $MULTIPORT $RULE2 $TARGETX"
        
        if [ $DO_IPTABLES = 1 ]; then
          # Outgoing traffic
          run $IPTABLES   -I "$CHAIN_EZ"                \
                          -t mangle                     \
                          -o $IF_INET                   \
                          $RULE1

          if [ "$MIRRORRULE" == "1" ]; then
            # Incoming traffic
            run $IPTABLES -I "$CHAIN_EZ"                \
                          -t mangle                     \
                          -i $IF_INET                   \
                          $RULE2
          fi
        fi
      fi
    done
    exec 3>&-                 # Close fd 3.
    
    [ $DEFAULTCLASS = 0 ] && DEFAULTCLASS=$[${#CLASSES[*]}-1]
  fi
}

####################
# tc commands
####################
tc_stop()
{
  run $TC qdisc del dev $IF_INET root
}

tc_start()
{
  local -i DEFAULT=${CLASSES[$DEFAULTCLASS]%% *}
  
  RESULT=`run $TC qdisc add dev $IF_INET root handle 1:0 $QDISC default $DEFAULT r2q 1 2>&1`

  if [ "$RESULT" != "" ]; then
    echo $RESULT >&2
    [ "${RESULT:0:13}" = "Unknown qdisc" ] && \
      error "Qdisc $QDISC probably not supported by this $TC version.\n\tCheck here: http://luxik.cdi.cz/~devik/qos/htb" >&2
  fi
  
  run $TC class add dev $IF_INET parent 1:0 classid 1:1 $QDISC rate $LINK_RATE
  
  local -i C=1
  while test true
  do
    CLASS=${CLASSES[$C]}
    [ "$CLASS" = "" ] && break;
    
    set $CLASS
    CLASSID=$1
    RATE=$2
    CEIL=$3
    PRIO=$[$CLASSID-$OFFSET_PRIO]
    
    # Add leaf classes
    run $TC class add dev $IF_INET parent 1:1   \
            classid 1:$CLASSID $QDISC           \
            rate $RATE                          \
            ceil $CEIL                          \
            burst 100 cburst 100                \
            prio $PRIO
    
#           quantum ${TC_QUAN[$PRIO]}           \
    
    # Filter traffic to this leaf
    run $TC filter add dev $IF_INET parent 1:0  \
            protocol ip prio $PRIO              \
            handle $CLASSID                     \
            fw flowid 1:$CLASSID
    
    C=$[$C+1]
  done
}

tc_status()
{
  echo
  $TC -s qdisc ls dev $IF_INET

  if [ $VERBOSE = 1 ]; then
    $TC -s class ls dev $IF_INET
  fi
  
  local QDISC=
  local CLASSID=
  local PARENT=
  local PRIO=
  local RATE=
  local CEIL=
  local BURST=
  local CBURST=
  local SENTB=
  local SENTP=
  local DROPPED=
  local OVERL=
  local BPS=
  local PPS=
  local BACKLOG=
  local LENT=
  local BORROW=
  local GIANTS=
  
  local SENTB_=
  local SENTP_=
    
  echo -e "\tclassid\t\t\tsent\tsent\tdropped\tover\t\tborrow"
  echo -e "parent\t (prio)\trate\tceil\tbytes\tpackets\tpackets\tlimits\tlent\t     ed"
  echo -e "-------\t-------\t-------\t-------\t-------\t-------\t-------\t-------\t-------\t-------"
  
  STATS=`$TC -s class ls dev $IF_INET`
  [ "$STATS" = "" ] && error No qdisc and classes defined in tc.
  set $STATS
  while test true
  do
    while [ "$1" != "" -a "$1" != "class" ]; do
      shift 1
    done
    [ "$1" = "" ] && break
    
    QDISC=$2
    CLASSID=$3
    if [ "$4" = "root" ]; then
      ROOT=1
      PARENT=-
      PRIO=""
    else
      ROOT=0
      CLASSID=${CLASSID#*:}
      PARENT=$5
      PRIO=$7
      shift 3         # NOTE !!
    fi
    RATE=${6/K/k}
    RATE=${RATE/bit/b}
    RATE=${RATE/byte/B}
    CEIL=${8/K/k}
    CEIL=${CEIL/bit/b}
    CEIL=${CEIL/byte/B}
    BURST=${10}
    CBURST=${12}
    SENTB=${14}
    SENTP=${16}
    DROPPED=${19%%,}
    OVERL=${21%%)}
    if [ "${22}" = "rate" ]; then
      BPS=${23}
      shift 2
      if [ "${22}" != "lended:" -a "${22}" != "backlog" ]; then
        PPS=${23}
        shift 1
      fi
    fi
    if [ "${22}" = "backlog" ]; then
      BACKLOG=${23}
      shift 2
    fi
    LENT=${23}
    BORROW=${25}
    GIANTS=${27}
    
    shift 30
    
    [ $DROPPED = 0  ] && DROPPED=
    [ $OVERL   = 0  ] && OVERL=
    [ "$PRIO" != "" ] && PRIO=" ($PRIO)"
    
    SENTB_=`addsuffix   $SENTB 1024`
    SENTP_=`addsuffix   $SENTP 1000`
    LENT_=`addsuffix    $LENT 1000`
    DROPPED_=`addsuffix $DROPPED 1000`
    OVERL_=`addsuffix   $OVERL 1000`
    BORROW_=`addsuffix  $BORROW 1000`
    GIANTS_=`addsuffix  $GIANTS 1000`
    
    echo -e "$PARENT\t$CLASSID$PRIO\t$RATE\t$CEIL\t$SENTB_\t$SENTP_\t$DROPPED_\t$OVERL_\t$LENT_\t$BORROW_"
  done | sort -n -
}

####################
# Main script
####################
defaults
parse_options $*
exit 0
